import zmq
import time
import sys
import logging
logging.basicConfig(level=logging.INFO)

from ztl.core.protocol import Message, Request, State

class TaskServer(object):

  def __init__(self, port):
    self.logger = logging.getLogger('remote-task')
    context = zmq.Context()
    self.socket = context.socket(zmq.REP)
    address = "tcp://*:" + str(port)
    self.socket.bind(address)
    self.controllers = {}
    self.logger.info("Task Server listening at '%s'" % address)


  def send_message(self, scope, mid, state, payload):
    self.socket.send(Message.encode(scope, mid, state, payload))


  def register(self, scope, controller):
    """
    Register a new controller for requests on a specific scope.

    Parameters
    ----------
    scope: The scope for which the server should dispatch tasks to the controller. Will replace any existing controller for this scope.
    controller: A controller object that will be called with requests (init, status, abort)
    """

    self.logger.info("Registering controller for scope '%s'." % scope)
    self.controllers[scope] = controller


  def unregister(self, scope):
    """
    Unregister any controller on a given scope.

    Parameters
    ----------
    scope: The scope for which the server should not dispatch tasks any longer.
    """
    self.controllers[scope] = None
    self.logger.info("Controller for scope '%s' removed." % scope)


  def listen(self):
    """
    Begin listening to requests and dispatching them to any controllers if available. This method blocks until interrupted.

    """
    while True:
      try:
        message = self.socket.recv()
        request = Message.decode(message)

        if all(field in request for field in Message.FIELDS):

          scope = request["scope"]
          if scope in self.controllers:
            controller = self.controllers[scope]

            state = int(request["state"])
            mid = int(request["id"])
            payload = request["payload"]

            try:
              if state == Request.INIT:
                ticket, response = controller.init(payload)
                if ticket > 0:
                  self.send_message(scope, State.ACCEPTED, ticket, response)
                else:
                  self.send_message(scope, State.REJECTED, ticket, response)
              elif state == Request.STATUS:
                status, response = controller.status(mid, payload)
                self.send_message(scope, status, mid, response)
              elif state == Request.ABORT:
                status, response = controller.abort(mid, payload)
                self.send_message(scope, status, mid, response)
              else:
                self.send_message(scope, State.REJECTED, mid, "Invalid request")
            except Exception as e:
              logging.error(e)
              self.send_message(scope, State.FAILED, mid, "Controller threw exception: " + str(e))
          else:
            self.send_message(scope, State.REJECTED, -1, "No controller for scope: " + scope)
            self.logger.warning("No controller for scope '%s', ignoring." % scope)

        else:
          self.send_message(scope, State.REJECTED, -1, "Unknown protocol")
          self.logger.warning("Unknown command received '%s', ignoring." % message)

      except Exception as e:
        logging.error(e)
        time.sleep(1)
      except KeyboardInterrupt:
        logging.error("Interrupted, exiting.")
        sys.exit(1)



import pytest

from unittest import TestCase

from ztl.core.client import RemoteTask
from ztl.core.server import TaskServer
from ztl.core.protocol import State

@pytest.mark.usefixtures("ztl_server")
class TestLocalLife(TestCase):

  def test_no_controller(self):
    host = "localhost"
    port = 7778
    scope = "/no"
    payload = "does not matter"

    task = RemoteTask(host, port, scope)
    mid, reply = task.trigger(payload)

    assert mid < 0

  def test_none_controller(self):
    host = "localhost"
    port = 7778
    scope = "/none"
    payload = "does not matter"

    task = RemoteTask(host, port, scope)
    mid, reply = task.trigger(payload)

    assert mid < 0

@pytest.mark.usefixtures("ztl_simple_server")
class TestLifeCycle(TestCase):

  def test_reject(self):
    host = "localhost"
    scope = "/test"
    port = 7777
    payload = "illegal payload"

    task = RemoteTask(host, port, scope)
    mid, reply = task.trigger(payload)

    assert mid < 0

  def test_abort(self):
    host = "localhost"
    scope = "/test"
    port = 7777
    payload = 5

    task = RemoteTask(host, port, scope)
    mid, reply = task.trigger(payload)

    assert mid > 0

    state, reply = task.wait(mid, .1)
    assert state == State.ACCEPTED

    state, reply = task.abort(mid)
    assert state == State.ABORTED

  def test_completion(self):
    host = "localhost"
    port = 7777
    scope = "/test"
    payload = 5

    task = RemoteTask(host, port, scope)
    mid, reply = task.trigger(payload)

    assert mid > 0

    state, reply = task.wait(mid, .1)
    assert state == State.ACCEPTED

    state, reply = task.wait(mid, 3.1)
    assert state == State.COMPLETED



import pytest

from unittest import TestCase

from ztl.core.client import RemoteTask
from ztl.core.server import TaskServer

class TestConstruction(TestCase):

  def test_server_construction(self):
    server = TaskServer(7779)
    server.register("/dummy", None)

  def test_client_construction(self):
    RemoteTask("localhost", 7779, "/dummy")



#!/usr/bin/env python

import time
import oyaml as yaml
import argparse
import os
import logging

logging.basicConfig(level=logging.INFO)

from sys import stdin, exit
from ztl.core.protocol import State, Task
from ztl.core.client import RemoteTask
from ztl.core.config import Remotes

class _Getch:
    """
    Gets a single character from standard input. Does not echo to the screen.
    """
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self): return self.impl()

class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()

class ScriptExecutor(object):

  tasks = {}

  def __init__(self, configfile, scriptfile):
    self.logger = logging.getLogger('script-exec')
    self.getch = _Getch()
    self.lastScene = None
    self.remotes = Remotes(configfile)

    with open(scriptfile) as f:
      self.script = yaml.safe_load(f)

  def parse_stage(self, stage):
    name = str(stage)
    delay = 0
    wait = True

    o = name.find("(")
    c = name.find(")")
    if o > 0 and c > 0:
      params = name[o+1:c]
      name = name[0:o]
      for param in params.split(","):
        pp = param.split("=")
        if len(pp) == 2:
          key = pp[0].strip()
          value = pp[1].strip()
          if key == "delay":
            delay = int(value)
          if key == "wait":
            wait = value.lower() in ['true', '1', 't', 'y', 'yes', 'on']
    return name, delay, wait

  def execute_scene(self, scene):

    scene_name, scene_delay, scene_wait = self.parse_stage(scene)
    steps = list(self.script[scene].keys())

    print(("EXECUTING SCENE '%s'" % scene_name) + (" WITH DELAY %ss" % scene_delay if scene_delay > 0 else "") + "...")

    if scene_delay > 0:
      time.sleep(scene_delay)

    for step in steps:
      step_name, step_delay, step_wait = self.parse_stage(step)
      print(("STARTING STEP '%s'" % step_name) + (" IN BACKGROUND" if not step_wait else "") + (" WITH DELAY %ss" % step_delay if step_delay > 0 else "") + "...")

      if step_delay > 0:
        time.sleep(step_delay)

      task_ids = []
      handlers = self.script[scene][step].keys()
      for handler in handlers:
        remote = self.remotes.get(handler)
        if not remote is None:
          components = self.script[scene][step][handler].keys()
          for component in components:
            component_name, component_delay, component_wait = self.parse_stage(component)
            goal = self.script[scene][step][handler][component]

            if component_delay > 0:
              time.sleep(component_delay)

            remote_id, reply = remote.trigger(Task.encode(handler, component_name, goal))
            if remote_id > 0:
              if step_wait and component_wait:
                task_ids.append(str(remote_id) + ":" + str(handler) + ":" + str(component_name) + ":" + str(goal))
              else:
                # MOVE TO DEBUG AFTER FINISHING THIS FEATURE
                self.logger.info("Component '%s' on handler '%s' for step '%s' TRIGGERED TO RUN IN BACKGROUND." % (component_name, handler, step_name))
            else:
              self.logger.error("Component '%s' on handler '%s' for step '%s' COULD NOT BE TRIGGERED. REPLY: '%s'" % (component_name, handler, step_name, reply))
        else:
          self.logger.error("No remote for handler '%s'. Step '%s' COULD NOT BE TRIGGERED. REPLY: '%s'" % (handler, step_name, reply))

      running = True
      while running:
        running = False
        for task_id in task_ids:
          components = task_id.split(":")
          remote_id = int(components[0])
          handler = components[1]
          status, reply = self.remotes.get(handler).wait(remote_id, task_id, timeout=100)
          running = running or status <= State.ACCEPTED
        time.sleep(.1)


  def confirm_scene(self, scene):
    print("\n----------------------------")
    print("ABOUT TO EXECUTE SCENE '%s'" % scene)

    steps = list(self.script[scene].keys())

    for step in steps:
      print("STEP: %s" % step)
      handlers = self.script[scene][step].keys()
      for handler in handlers:
        components = self.script[scene][step][handler].keys()
        for component in components:
          goal = self.script[scene][step][handler][component]
          print("\t%s [%s]: -> %s" % (handler, component, goal))

    if self.lastScene == None:
      print("PRESS <ENTER> TO CONFIRM or ANY OTHER KEY TO SKIP")
      return self.get_key()
    else:
      print("PRESS <ENTER> TO CONFIRM, PRESS <R> TO REPLAY LAST SCENE OR ANY OTHER KEY TO SKIP")
      return self.get_key()

  def get_key(self):
    first_char = self.getch()
    if first_char == '\x03':
      exit(1)
    # The idea would be to allow further decomposition of the getch e.g. if arrows keys are pressed
    return first_char

  def execute(self):
    try:
        restart = True
        while restart:
          for scene in self.script.keys():
            repeat = True
            while repeat:
              keyPressed = self.confirm_scene(scene)
              if keyPressed == "\r" or keyPressed == b"\r":
                self.execute_scene(scene)
                self.lastScene = scene
                repeat = False
              elif self.lastScene != None and (keyPressed == "r" or keyPressed == "R" or keyPressed == b"r" or keyPressed == b"R"):
                print("\n Repeating Scene '%s" % (self.lastScene))
                self.execute_scene(self.lastScene)
              else:
                print("\nSKIPPING SCENE '%s'" % (scene))
                repeat = False
          if self.lastScene == None:
            print("PRESS <A> TO RESTART THE SCRIPT OR ANY OTHER KEY TO EXIT")
            keyPressed = self.get_key()
            if keyPressed == "a" or keyPressed == "A" or keyPressed == b"a" or keyPressed == b"A":
              print("\n Repeating Script")
              restart = True
            else:
              restart = False
          else:
            repeat = True
            while repeat:
              print("PRESS <R> TO REPLAY THE LAST SCENE, PRESS <A> TO RESTART THE SCRIPT OR ANY OTHER KEY TO EXIT")
              keyPressed = self.get_key()
              if self.lastScene != None and (keyPressed == "r" or keyPressed == "R" or keyPressed == b"r" or keyPressed == b"R"):
                    print("\n Repeating Scene '%s" % (self.lastScene))
                    self.execute_scene(self.lastScene)
              elif keyPressed == "a" or keyPressed == "A" or keyPressed == b"a" or keyPressed == b"A":
                    print("\n Repeating Script")
                    repeat = False
                    restart = True
              else:
                repeat = False
                restart = False
          self.lastScene = None
    except Exception as e:
      self.logger.error(e)
      exit(1)
    except KeyboardInterrupt:
      self.logger.error("Interrupted, exiting.")
      exit(1)

def main_cli():

  cfg_file = os.environ.get('XDG_CONFIG_HOME', os.environ.get('HOME', '/home/demo') + '/.config') + '/zmq-remotes.yaml'

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument("-c", "--config", type=str,
                      help="Configuration file location.", default=cfg_file)
  parser.add_argument("-s", "--script", type=str,
                      help="Script file to execute.", required=True)

  args, unknown = parser.parse_known_args()
  run = ScriptExecutor(args.config, args.script)
  run.execute()


if __name__ == "__main__":

  main_cli()




from unittest import TestCase

class TestUnitTests(TestCase):

    def test_unit_testing(self):
        self.assertTrue(True)

class TestImports(TestCase):

    def test_zmq_import(self):
        import zmq

    def test_ztl_protocol_imports(self):
        from ztl.core.protocol import Message, Request, State, Task

    def test_ztl_server_imports(self):
        from ztl.core.server import TaskServer

    def test_ztl_client_imports(self):
        from ztl.core.client import RemoteTask

    def test_ztl_task_imports(self):
        from ztl.core.task import ExecutableTask, TaskController, TaskExecutor

    def test_ztl_config_imports(self):
        from ztl.core.config import Remotes




import yaml

import logging
logging.basicConfig(level=logging.INFO)

from ztl.core.client import RemoteTask

class Remotes():

  def __init__(self, configfile):

    self.logger = logging.getLogger('remote-config')
    self.remotes = {}
    self.config = {}

    with open(configfile) as f:
      self.config = yaml.safe_load(f)


  def get(self, name):

    if name in self.remotes:
      return self.remotes[name]

    else:
      rs = self.config["remotes"]
      if name in rs.keys():
        self.add(name, rs[name]["host"], rs[name]["port"], rs[name]["scope"])
        return self.remotes[name]

      return None


  def has(self, name):
    return name in self.remotes


  def add(self, name, host, port, scope):
    self.logger.info("Initialising remote task interface '%s'..." % name)
    self.remotes[name] = RemoteTask(host, port, scope)



#!/usr/bin/env python

import sys
import time

import argparse

from ztl.core.server import TaskServer
from ztl.core.protocol import State, Task
from ztl.core.task import ExecutableTask, TaskExecutor, TaskController


class TestTask(ExecutableTask):

  def __init__(self, component, goal):
    self.active = True
    self.component = component
    self.goal = goal

  def execute(self):
    if self.component == "echo":
      return self.goal
    if self.component == "print":
      print("component: " + self.component)
      print("goal: " + self.goal)
      return True
    if self.component == "sleep":
      time.sleep(int(self.goal))
      return True
    else:
      raise RuntimeError("No such component '%s', try 'echo/print/sleep'." % self.component)

  def abort(self):
    return False


class TestController(TaskController):

  def assign(self, handler, component, goal):
    if handler == "test":
      return TestTask, component, goal
    raise RuntimeError("No such handler '%s', try 'test'." % handler)


def main_cli():

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument("-p", "--port", type=int,
                      help="The port on the local machine that the server should listen to.", required=True)
  parser.add_argument("-s", "--scope", type=str,
                      help="The scope that the server should respond to.", required=True)

  args, unknown = parser.parse_known_args()

  server = TaskServer(args.port)
  server.register(args.scope, TestController())
  server.listen()


if __name__ == "__main__":

  main_cli()



import base64

class Request:

  INIT = 1
  STATUS = 2
  ABORT = 3

  @staticmethod
  def name(code):
    if code == Request.INIT: return "INIT"
    if code == Request.STATUS: return "STATUS"
    if code == Request.ABORT: return "ABORT"
    return None

class State:

  INITIATED = 0
  ACCEPTED = 1
  REJECTED = 2
  FAILED = 3
  ABORTED = 4
  COMPLETED = 5

  @staticmethod
  def name(code):
    if code == State.INITIATED: return "INITIATED"
    if code == State.ACCEPTED: return "ACCEPTED"
    if code == State.REJECTED: return "REJECTED"
    if code == State.FAILED: return "FAILED"
    if code == State.ABORTED: return "ABORTED"
    if code == State.COMPLETED: return "COMPLETED"
    return None

class Message:

  SEPARATOR = ";"
  FIELDS = ["scope", "state", "id", "payload"]

  @staticmethod
  def encode(scope, state, mid, payload):
    msg = {"scope": str(scope),
           "state": str(state),
           "id": str(mid),
           "payload": str(payload)}
    return (msg["scope"] + Message.SEPARATOR  + msg["state"] + Message.SEPARATOR + msg["id"] + Message.SEPARATOR + msg["payload"]).encode('utf-8')

  @staticmethod
  def decode(message):
    split = message.decode("utf-8").split(Message.SEPARATOR, len(Message.FIELDS) - 1)
    unfolded = dict(zip(Message.FIELDS, split))
    return unfolded


class Task:

  SEPARATOR = ":"
  FIELDS = ["handler", "component", "goal"]

  @staticmethod
  def decode(message):
    cmd = base64.b64decode(bytes(str(message).encode("utf-8"))).decode("utf-8").split(Task.SEPARATOR, len(Task.FIELDS) - 1)
    return dict(zip(Task.FIELDS, cmd))

  @staticmethod
  def encode(handler, component, goal):
    joined = str(handler) + Task.SEPARATOR + str(component) + Task.SEPARATOR + str(goal)
    code = base64.b64encode(joined.encode('utf-8'))
    return code.decode("utf-8")



import zmq
import time
import sys
import logging
logging.basicConfig(level=logging.INFO)

from ztl.core.protocol import Message, Request, State

class RemoteTask(object):

  def __init__(self, host, port, scope, timeout=2000):

    self.logger = logging.getLogger('remote-task')
    self.context = zmq.Context()
    self.socket = self.context.socket(zmq.REQ)
    self.socket.setsockopt(zmq.RCVTIMEO, timeout)
    self.socket.setsockopt(zmq.LINGER, 1)
    self.address = "tcp://" + str(host) + ":" + str(port)
    self.socket.connect(self.address)
    self.scope = scope

    self.logger.info("Remote task interface initialised at '%s'.", self.address)

  def trigger(self, payload):
    """
    Trigger a task at the remote interface.

    Parameters
    ----------
    payload: The payload containing the task description.

    Returns
    -------
    id: An ID for the task assigned by the server if accepted, -1 if rejected or server not reachable or communication error.
    reply: The remote reply containing an updated task description or the error message if rejected or server not reachable or communication error.
    """
    msg = Message.encode(self.scope, Request.INIT, -1, payload)
    try:
      self.socket.send(msg)
      reply = Message.decode(self.socket.recv())
      return int(reply["id"]), reply["payload"]
    except Exception as e:
      self.logger.error(repr(e))
      return -1, repr(e)

  def abort(self, mid, payload="abort command"):
    """
    Aborts a task at the remote interface.

    Parameters
    ----------
    mid: The ID of the task to abort.
    payload: An optional payload containing an updated task description. Default: "abort command"

    Returns
    -------
    status: The updated status after attempting to abort. Task might not be aborted. -1 if server not reachable or communication error.
    reply: The remote reply containing an updated task description. May contain error message if task not aborted or server not reachable or communication error.
    """
    try:
      msg = Message.encode(self.scope, Request.ABORT, mid, payload)
      self.socket.send(msg)
      reply = Message.decode(self.socket.recv())
      return int(reply["state"]), reply["payload"]
    except Exception as e:
      self.logger.error(repr(e))
      return -1, repr(e)

  def status(self, mid, payload="status update"):
    """
    Query the status of a given task at the remote interface.

    Parameters
    ----------
    mid: The ID of the task to query about.
    payload: An optional payload containing an updated task description. Default: "status update"

    Returns
    -------
    status: The current status of the task at the server. State.FAILED if server not reachable or communication error.
    reply: The remote reply containing an updated task description or the error message if server not reachable or communication error.
    """
    try:
      msg = Message.encode(self.scope, Request.STATUS, mid, payload)
      self.socket.send(msg)
      reply = Message.decode(self.socket.recv())
      return int(reply["state"]), reply["payload"]
    except Exception as e:
      self.logger.error(repr(e))
      return State.FAILED, repr(e)

  def wait(self, mid, payload = "waiting poll", timeout = 5.0):
    """
    Wait for a task to complete at the remote interface.

    Parameters
    ----------
    mid: The ID of the task to wait for.
    payload: An optional payload containing an updated task description. Default: "waiting poll"
    timeout: Maximum time to wait, interpreted as infinite when equal or smaller than 0. Default: 5.0 secs.

    Returns
    -------
    state: The last task status after waiting complete. State.FAILED if server not reachable or communication error.
    reply: The remote reply containing an updated task description or the error message if server not reachable or communication error.
    """
    start = time.time()
    state = -1
    reply = None
    while timeout <= 0 or (time.time() - start) < timeout:
      state, reply = self.status(mid, payload = payload)
      if state > State.ACCEPTED:
        return state, reply
      time.sleep(.1)
    return state, reply



#!/usr/bin/env python

import os
import sys

import argparse

from ztl.core.client import RemoteTask
from ztl.core.protocol import State, Task
from ztl.core.config import Remotes

def main_cli():

  cfg_file = os.environ.get('XDG_CONFIG_HOME', os.environ.get('HOME', '/home/demo') + '/.config') + '/zmq-remotes.yaml'

  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument("-c", "--config", type=str,
                      help="Configuration file location.", default=cfg_file)
  parser.add_argument("-t", "--timeout", type=int,
                      help="An optional timeout to wait for task completion, continues waiting if 0 or less", required=False, default=-1)
  parser.add_argument("remote", type=str,
                      help="The task remote as specified in the configuration file.")
  parser.add_argument("payload", type=str,
                      help="The task specification payload, with handler, component and goal separated by ':', e.g. 'handler:component:goal'.")


  args, unknown = parser.parse_known_args()
  cmd = args.payload.split(":", 2)
  if len(cmd) < 3:
    print("Error: Invalid task specification. Payload needs to specify handler, component and goal separated by ':', e.g. 'handler:component:goal'")
    parser.print_help(sys.stderr)
    return -1

  remotes = Remotes(args.config)
  task = remotes.get(args.remote)

  request = Task.encode(cmd[0], cmd[1], cmd[2])

  print("Triggering task with request '%s'..." % request)
  mid, reply = task.trigger(request)

  if mid > 0:
    print("Initialised task with ID '%s' for %ss. Reply is '%s'." % (mid, args.timeout, reply))
    state, reply = task.wait(mid, timeout=args.timeout)
    if state < 0:
      print("Could not wait for task with ID '%s'. Reply is '%s'." % (mid, reply))
    elif state <= State.ACCEPTED:
      print("Aborting task with ID '%s'..." % mid)
      state, reply = task.abort(mid)
      if state == State.ABORTED:
        print("Task with ID '%s' aborted. Reply is '%s'." % (mid, reply))
      elif state <= State.ACCEPTED:
        print("Could not abort Task with ID '%s', waiting for completion. Reply is '%s'." % (mid, reply))
        state, reply = task.wait(mid)
        print("Task with ID '%s' finished in state '%s' after unsuccessful abort signal. Reply is '%s'." % (mid, State.name(state), reply))
    else:
      print("Task with ID '%s' finished in state '%s' while waiting. Result is '%s'." % (mid, State.name(state), reply))

  else:
    print("Task could not be triggered: '%s'." % reply)

if __name__ == "__main__":

  main_cli()
